"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SpreadMatcher = void 0;
const lodash_1 = require("lodash");
const bignumber_1 = require("@ethersproject/bignumber");
var SpreadMatcher;
(function (SpreadMatcher) {
    SpreadMatcher.spreadMatcher = (originalData) => {
        const data = lodash_1.cloneDeep(originalData);
        const allRoundMatches = [];
        let i = 0;
        while ((i += 1)) {
            if (i > 10000) {
                throw new Error(`
                  Matching executed more than 10000 - it may mean, that something broke
                  data: ${JSON.stringify(data)}
                  matches: not displayed, because of potential length
                `);
            }
            const allGroupsHaveVolume = [
                data.entityGroups[0].some((v) => v.volume.gt(0)),
                data.entityGroups[1].some((v) => v.volume.gt(0))
            ].every(Boolean);
            if (!allGroupsHaveVolume) {
                break;
            }
            const roundInput = getRoundInput(data);
            if (roundInput.routes.length === 0) {
                break;
            }
            const roundMatches = matchRound(roundInput);
            roundMatches.forEach((match) => {
                const entity1 = data.entityGroups[0].find((e) => e === match.entities[0]);
                const entity2 = data.entityGroups[1].find((e) => e === match.entities[1]);
                entity1.volume = entity1.volume.sub(match.volume);
                entity2.volume = entity2.volume.sub(match.volume);
            });
            allRoundMatches.push(roundMatches);
        }
        return {
            matches: sumMatches(allRoundMatches.flat()).map(omitMatchVolumes),
            roundMatches: allRoundMatches,
            leftoverEntities: data.entityGroups
        };
    };
    const omitMatchVolumes = (match) => {
        return {
            ...match,
            entities: match.entities.map((e) => lodash_1.omit(e, 'volume'))
        };
    };
    const sumMatches = (matches) => {
        const areMatchesSame = (match1, match2) => {
            return (match1.entities[0] === match2.entities[0] &&
                match1.entities[1] === match2.entities[1]);
        };
        return matches.reduce((sum, match) => {
            const existingMatchIndex = sum.findIndex((s) => areMatchesSame(s, match));
            if (existingMatchIndex >= 0) {
                return sum.map((m, i) => {
                    if (i === existingMatchIndex) {
                        return {
                            ...m,
                            volume: m.volume.add(match.volume)
                        };
                    }
                    else {
                        return m;
                    }
                });
            }
            return [...sum, match];
        }, []);
    };
    const computeSafeDistribution = (entities1, entities2, routes) => {
        const computeDistributedVolume = (entities) => entities.map((entity) => ({
            entity,
            volume: entity.volume.div(getCompetingEntities(routes, entity).length)
        }));
        return bigNumMinBy([...computeDistributedVolume(entities1), ...computeDistributedVolume(entities2)], (e) => e.volume);
    };
    const getCompetingEntities = (routes, entity) => {
        return routes
            .filter((route) => route.includes(entity))
            .flat()
            .filter((competingEntity) => competingEntity !== entity);
    };
    const matchRound = (input) => {
        const toDistribute = computeSafeDistribution(input.entityGroups[0], input.entityGroups[1], input.routes);
        if (toDistribute.volume.eq(0)) {
            const competingEntities = getCompetingEntities(input.routes, toDistribute.entity);
            const isEntityFirst = input.entityGroups[0].find((e) => e === toDistribute.entity);
            const entitiesToSatisfy = competingEntities.slice(0, toDistribute.entity.volume.toNumber());
            return entitiesToSatisfy.map((connectedEntity) => ({
                entities: (isEntityFirst
                    ? [toDistribute.entity, connectedEntity]
                    : [connectedEntity, toDistribute.entity]),
                volume: bignumber_1.BigNumber.from(1)
            }));
        }
        return input.routes.map((route) => ({
            entities: route,
            volume: toDistribute.volume
        }));
    };
    const getRoundInput = (data) => {
        const entity1Group = getFirstUsableGroup(data.groupPriority, data.entityGroups[0]);
        const routes = entity1Group.flatMap((entity1GroupEntry) => {
            const preferredEntity2Group = getFirstUsableGroup(entity1GroupEntry.groupPriority, data.entityGroups[1]);
            return cartesian(entity1GroupEntry.entities, preferredEntity2Group.flatMap((g) => g.entities));
        });
        return {
            routes,
            entityGroups: [
                data.entityGroups[0].filter((entity) => routes.some((e) => e[0].id === entity.id)),
                data.entityGroups[1].filter((entity) => routes.some((e) => e[1].id === entity.id))
            ]
        };
    };
    const getFirstUsableGroup = (groups, entities) => {
        const entitiesWithVolume = entities.filter((e) => e.volume.gt(0));
        const groupsWithVolume = groups
            .map((group) => {
            const entitiesInGroup = group.map((entry) => ({
                ...entry,
                entities: entitiesWithVolume.filter((e) => e.id === entry.id)
            }));
            return entitiesInGroup;
        })
            .filter((group) => group.some((g) => g.entities.length > 0));
        return groupsWithVolume[0] ?? [];
    };
    const bigNumMinBy = (collection, cb) => {
        return collection.reduce((smallest, current) => {
            return cb(current).lt(cb(smallest)) ? current : smallest;
        }, collection[0]);
    };
    const cartesian = (arr1, arr2) => {
        return arr1.flatMap((val1) => arr2.map((val2) => [val1, val2]));
    };
})(SpreadMatcher = exports.SpreadMatcher || (exports.SpreadMatcher = {}));
//# sourceMappingURL=spreadMatcher.js.map